#!/usr/bin/python2
# coding=utf-8
import logging
from argparse import FileType


class BioFormat(object):
    """
    Returns tuples (ID, sequence_length, sequence) from CSV and FASTA files
    """
    def __init__(self, file_handle, logger_name="root"):
        if not (isinstance(file_handle, file) or isinstance(file_handle, FileType)):
            raise TypeError("BioFormat needs a file handle, given instance of {}".format(type(file_handle)))
        self.file_handle = file_handle
        self.logger = logging.getLogger(logger_name)
        input_type = "CSV"
        has_header = False
        with open(file_handle.name, "r") as infile:
            stripped_line = infile.readline().strip()
            if stripped_line.startswith('>'):
                input_type = "FASTA"
            elif stripped_line.count(';') == 2:
                split_line = stripped_line.split(';')
                if split_line[0] == "ID":
                    has_header = True
                else:
                    try:
                        int(split_line[1])
                    except ValueError as e:
                        self.logger.exception("This is not a proper csv file")
                        raise

            else:
                raise ValueError("Wrong file type")
        self.type = input_type.upper()
        self.has_header = has_header

        self.logger.info("Detected {} input file {} header".format(self.type, "with" if self.has_header else "without"))
        if self.type == "CSV":
            import csv
            self.iterator = csv.reader(self.file_handle, delimiter=';')
            if has_header:
                header = self.iterator.next()
                if header[0] != "ID":
                    raise ValueError("Wrong csv header:\n{}".format(header[0]))
            self.get_record = self._get_csv
        elif self.type == "FASTA":
            from Bio import SeqIO
            self.iterator = SeqIO.FastaIO.FastaIterator(self.file_handle)
            self.get_record = self._get_fasta

    def _get_fasta(self):
        for fdata in self.iterator:
            seq = str(fdata.seq)
            yield (fdata.id, len(seq), seq)

    def _get_csv(self):
        for row in self.iterator:
            yield (row[0], int(row[1]), row[2])
