#!/usr/bin/python2
# coding=utf-8

import errno
import gc
import hashlib
import logging
import math as m
import os
import shutil
import subprocess as sp
import sys
import traceback
from argparse import ArgumentParser, FileType
from collections import Counter
from functools import partial
from itertools import izip

import numpy as np
import psutil

from bioformat import BioFormat

have_scoop = None
parent = psutil.Process(os.getppid())

saved_fields = ['uid', 'binpath', 'file_size', 'seqlen', 'npoints', 'nuniq', 'window_size',
                'npoints_per_res', 'black_score', 'log_score']

if any(key == "scoop" or key == "scoop.launch.__main__" for key in parent.cmdline()):
    import scoop
    from scoop import futures

    have_scoop = True
else:
    have_scoop = False


def create_dir(path):
    try:
        os.mkdir(path)
    except OSError as e:
        if e.errno == errno.EEXIST:
            pass
        else:
            raise


def recursive_hex_create(path, nlevels):
    for hexdir in xrange(16):
        newpath = os.path.join(path, "{:x}".format(hexdir))
        create_dir(newpath)
        if nlevels > 1:
            recursive_hex_create(newpath, nlevels - 1)


def create_binpath_needed(uid, depth, outformat):
    outpath = list(hashlib.md5(uid).hexdigest()[:depth])
    outpath.append("{}.{}".format(uid, outformat))
    return os.path.join("individual-bin", *outpath)


def create_binpath_empty(uid, depth, outformat):
    return "--"


def rundotter(seqpacket, have_scoop, outformat="bin", print_only=False, depth=3, dotterpath=None, overwrite=False,
              threshold=0):
    import tempfile as tmp
    if have_scoop:
        logger = scoop.logger
    else:
        logger = logging.getLogger('dotrun')
    packet_size = len(seqpacket)
    warnings = {}
    a_uid = []
    a_status = np.zeros(shape=packet_size, dtype=np.uint8)
    a_existence = np.zeros(shape=packet_size, dtype=np.bool)
    a_file_size = np.zeros(shape=packet_size, dtype=np.uint64)
    a_seqlen = np.zeros(shape=packet_size, dtype=np.uint16)
    a_nuniq = np.zeros(shape=packet_size, dtype=np.uint16)
    a_npoints = np.zeros(shape=packet_size, dtype=np.uint32)
    a_window_size = np.zeros(shape=packet_size, dtype=np.uint8)
    a_npoints_per_res = np.zeros(shape=packet_size, dtype=np.float32)
    a_black_score = np.zeros(shape=packet_size, dtype=np.float32)
    a_log_score = np.zeros(shape=packet_size, dtype=np.float32)

    local_widths = Counter()

    create_binpath = create_binpath_needed

    const_dotter_opts = ['--horizontal-type=p', '--vertical-type=p', '-z', '1']

    if outformat == "bin":
        dotopt = "-k"
    elif outformat == "json":
        dotopt = "-j"
        if print_only:
            const_dotter_opts.append('-J')
            create_binpath = create_binpath_empty
    else:
        logger.exception(
            "The only allowed values for output format are: 'bin' and 'json', given {}".format(outformat))

    worker = None
    if have_scoop:
        worker = scoop.worker

    with tmp.NamedTemporaryFile(mode='w+t', prefix="dotrun") as fasta_file, \
            tmp.NamedTemporaryFile(mode='w+t', prefix="dotrun_err") as dotter_log:
        tname = fasta_file.name
        ename = dotter_log.name

        for i, seqdata in enumerate(seqpacket):
            seqid, seq = seqdata.split(":")
            seqlen = 0
            uid = "NONEXIST"
            status = 255
            existed = False
            npoints = 0
            nuniq = 0
            window_size = 0
            npoints_per_res = 0.
            black_score = 0.
            log_score = 0.

            fasta_file.seek(0)
            fasta_file.truncate()
            dotter_log.seek(0)
            dotter_log.truncate()

            try:
                existed = False
                if seqid.startswith("UniRef"):
                    uid = "{}".format(seqid.split('_')[1])
                else:
                    uid = "{}".format(seqid)

                binpath = create_binpath_needed(uid, depth, outformat)
                existed = os.path.exists(binpath)
                seqlen = len(seq)
                if not existed or overwrite:
                    # lst.write("{}\n".format(uid))

                    fasta_file.write(">{} seqlen={}\n{}\n".format(uid, seqlen, seq))
                    fasta_file.flush()

                    if worker:
                        dotter_log.write("Worker: {}".format(worker))
                    bpoint = str(threshold)
                    wpoint = str(threshold - 1)
                    dotcmd = [dotterpath, tname, tname, dotopt, binpath, '-x', bpoint, '-y', wpoint] + const_dotter_opts
                    # import socket
                    # print("Cmd for {} ({}): {}".format(uid, socket.gethostname(), dotcmd))
                    status = sp.call(dotcmd, stdout=dotter_log.file, stderr=sp.STDOUT)
                    dotter_log.flush()
                    dotter_log.seek(0)

                    out_vars = {}
                    if status == 0:
                        # parse output
                        for l in dotter_log.file:
                            if l.startswith("VARS"):
                                ldata = l.strip().split()[1:]
                                for item in ldata:
                                    key, value = item.split('=')
                                    out_vars[key] = value
                                break
                            elif l.startswith("WARNING"):
                                warnings[uid] = l[9:]
                                continue
                        else:
                            try:
                                copy_to_workdir(ename, tname, uid)
                            except IOError as e:
                                logging.warning("Could not copy files for sequence {} to {}/fasta "
                                                "due to IOError#{} {}.".format(uid, os.getcwd(), e.errno, e.message))
                            logging.warning("Dotter for {} finished with status 0 "
                                            "but produced no results".format(uid))

                        try:
                            nuniq = int(out_vars["nUniqueOffsets"])
                            npoints = int(out_vars["nPoints"])
                            log_score = float(out_vars["logScore"])
                            black_score = float(out_vars["blackScore"])
                            npoints_per_res = float(out_vars["nPointsPerResidue"])
                            window_size = int(out_vars["windowSize"])
                        except KeyError as e:
                            logger.warning("Could not find some variables in dotter output "
                                           "for sequence {}: {}".format(uid, e.message))
                    else:
                        try:
                            copy_to_workdir(ename, tname, uid)
                        except IOError as e:
                            logging.warning("Could not copy files for sequence {} to {}/fasta "
                                            "due to IOError#{} {}. Dotter call status {}".format(uid, os.getcwd(),
                                                                                                 e.errno, e.message,
                                                                                                 status))

                else:
                    status = 0
                    logger.info("File {} already exists, skipping...".format(binpath))
                    existed = True

            except Exception:
                etype, value, tb = sys.exc_info()
                lines = traceback.format_exception(etype, value, tb)

                print
                logger.info('---------------------- INTERNAL TRACEBACK ----------------------')
                print
                logger.info(seqid)
                logger.info(seq)
                logger.info(''.join(lines))
                logger.info('------------------ END OF INTERNAL TRACEBACK -------------------')
                print
                logger.exception("Error occured; please analyze the traceback above")

            try:
                file_size = os.path.getsize(binpath)
            except OSError:
                file_size = 0

            binpath = create_binpath(uid, depth, outformat)

            result_values = (uid, binpath, file_size, seqlen, npoints, nuniq, window_size,
                             npoints_per_res, black_score, log_score)

            if status == 0:
                if not existed or overwrite or print_only:
                    job_widths = Counter()
                    for key, value in izip(saved_fields, result_values):
                        if isinstance(value, str):
                            job_widths[key] = len(value)
                        elif isinstance(value, int):
                            if value > 0:
                                job_widths[key] = int(m.ceil(m.log10(value)))
                        elif isinstance(value, float):
                            if value > 0:
                                job_widths[key] = int(m.ceil(m.fabs(m.log10(value)))) + 3
                        else:
                            raise ValueError("Variable with key {} has unexpected "
                                             "type of {} with value {}".format(key, type(value), value))

                    local_widths |= job_widths
                else:
                    os.remove(binpath)

            a_uid.append(uid)
            a_status[i] = status
            a_existence[i] = existed
            a_file_size[i] = file_size
            a_seqlen[i] = seqlen
            a_npoints[i] = npoints
            a_nuniq[i] = nuniq
            a_window_size[i] = window_size
            a_npoints_per_res[i] = npoints_per_res
            a_black_score[i] = black_score
            a_log_score[i] = log_score

    return (a_uid, a_status, a_existence, a_file_size, a_seqlen, a_npoints, a_nuniq, a_window_size, a_npoints_per_res,
            a_black_score, a_log_score), worker, local_widths, warnings


def copy_to_workdir(ename, tname, uid):
    shutil.copy(tname, os.path.join(os.getcwd(), "fasta", "{}.fasta".format(uid)))
    shutil.copy(ename, os.path.join(os.getcwd(), "fasta", "{}.log".format(uid)))


def main():
    global have_scoop
    parser = ArgumentParser(description="A program to run dotter in a distributed environment")
    parser.add_argument("-f", "--input-file", type=FileType('r'), dest="infile", required=True,
                        help="A multi-sequence Fasta or CSV file. In CSV file, "
                             "destined to be used with large databases, each record should consist of "
                             "sequence id, sequence length and sequence itself")
    parser.add_argument("-d", "--dotter-path", type=str, dest="dotter_path", default="dotter",
                        help="Non-standard path to dotter")
    parser.add_argument("-t", "--threshold", type=int, dest="threshold", required=True,
                        help="Dot plot point value used for binarization, i.e. all values below become 0, "
                             "and all greater or equal become 1")
    parser.add_argument("-l", "--logfile", dest="logfile", default="dotrun.log", help="Log file name")
    parser.add_argument("-o", "--bin-list", type=str, dest="bin_list", default="single-bin.txt",
                        help="List of individual binary files within individual-bin folder")
    # parser.add_argument("-k", "--keep-if-repeat-protein", type=float, default=0.,
    #                     help="Keep only if number of points per residue is greater or equal than this value")
    parser.add_argument("-w", "--overwrite", action="store_true", dest="overwrite", default=False,
                        help="Overwrite existing files")
    parser.add_argument("-F", "--output-format", dest="output_format", type=str, choices=["json", "bin"], required=True,
                        help="Format of output: either binary ready for merging or JSON with statistics")
    parser.add_argument("-J", "--print-only", dest="print_only", action="store_true", default=False,
                        help="Do not write Json, only allow dotter to print the summary to the screen")
    parser.add_argument("-c", "--depth", dest="depth", required=True, type=int, choices=list(range(16)),
                        help="Create directory tree of this depth to limit the number of created files "
                             "per single directory; note that each level creates 16 subdirectories, so please adjust"
                             "this value so that number_of_sequences/16**depth is at most several thousand")
    parser.add_argument("-s", "--min-sequence-length", type=int, default=0, dest="min_sequence_length",
                        help="Minimum sequence length required for further processing")
    parser.add_argument("--filter", type=FileType('r'),
                        help="List of sequence UIDs to keep; everything else will be rejected")

    opts = parser.parse_args()

    if os.path.exists(opts.logfile):
        i = 1
        while os.path.exists("{}.{}".format(opts.logfile, i)):
            i += 1
        os.rename(opts.logfile, "{}.{}".format(opts.logfile, i))

    logger = logging.getLogger("dotrun")
    logger.setLevel(logging.DEBUG)
    fh = logging.FileHandler(opts.logfile, 'w')
    ch = logging.StreamHandler()
    ch.setLevel(logging.ERROR)
    # create formatter and add it to the handlers
    formatter = logging.Formatter('[%(asctime)-15s] %(name)-9s %(levelname)-7s %(message)s')
    fh.setFormatter(formatter)
    ch.setFormatter(formatter)
    # add the handlers to the logger
    logger.addHandler(fh)
    logger.addHandler(ch)

    logger.info("Args used:\n{}".format(" ".join(sys.argv)))

    if opts.print_only and opts.output_format.lower() == "bin":
        logger.exception("Option --print-only should only be used with json output format")

    nworkers = 1
    if have_scoop:
        logger.info("Working with SCOOP enabled")
        nworkers = scoop.SIZE
    else:
        logger.warn("WARN: This program runs sequentially on one core. This is likely not what you want")

    sequences_to_process = None
    if opts.filter:
        sequences_to_process = set()
        for line in opts.filter:
            l = line.strip()
            if l == "":
                continue
            sequences_to_process.add(l)
        logger.info("Read filter configuration: {} UIDs to search for".format(len(sequences_to_process)))

    proc = psutil.Process(os.getpid())
    gc.collect()

    logger.debug("Mem usage: start: {}".format(proc.memory_info().rss))

    npackets = 1
    if nworkers > 1:
        npackets = nworkers * 8
        # arbitrary constant 8

    # will send data in packets so that every worker has quite a few sequences to process
    fasta_data = [[] for _ in xrange(npackets)]
    ntoo_short = 0
    nprocessed = 0
    nsequences = 0
    b = BioFormat(opts.infile, "dotrun")
    for row in b.get_record():
        seqid, seqlen, seq = row
        if seqlen >= opts.min_sequence_length:
            if sequences_to_process is not None:
                if seqid not in sequences_to_process:
                    continue
            fasta_data[nsequences % npackets].append(":".join([seqid, seq]))
            if sequences_to_process:
                sequences_to_process.remove(seqid)
            nsequences += 1
        else:
            ntoo_short += 1
        nprocessed += 1
        if nprocessed % 1e6 == 0:
            logger.debug("Processed: {}, root mem usage: {}".format(nprocessed, proc.memory_info().rss))

    logger.info("Processed: {}, will run dotter on {} sequences".format(nprocessed, nsequences))
    if ntoo_short > 0:
        logger.info("Detected {} sequences below {} amino acids".format(ntoo_short, opts.min_sequence_length))

    if nsequences:
        if os.path.exists(opts.bin_list):
            i = 1
            while os.path.exists("{}.{}".format(opts.bin_list, i)):
                i += 1
            os.rename(opts.bin_list, "{}.{}".format(opts.bin_list, i))

        create_dir("fasta")
        create_dir("individual-bin")
        recursive_hex_create("individual-bin", opts.depth)
        rundotter_part = partial(rundotter, outformat=opts.output_format, print_only=opts.print_only,
                                 have_scoop=have_scoop,
                                 dotterpath=opts.dotter_path, overwrite=opts.overwrite,
                                 threshold=opts.threshold, depth=opts.depth)

        logger.debug("Mem usage: input read: {}".format(proc.memory_info().rss))

        if have_scoop:
            my_map = futures.map_as_completed
        else:
            my_map = map

        total_file_size = 0
        kept_file_size = 0
        kept_files = 0
        global_widths = Counter()

        for f in saved_fields:
            global_widths[f] = len(f)

        saved_results = []
        nprocessed = 0
        global_warnings = {}

        create_binpath = create_binpath_needed
        if opts.print_only:
            create_binpath = create_binpath_empty

        for packet_result in my_map(rundotter_part, fasta_data):
            arrays, worker, local_widths, warnings = packet_result
            global_warnings.update(warnings)
            global_widths |= local_widths
            saved_results.append(arrays)
            nprocessed += 1
            if have_scoop:
                logger.info("Processed {}/{} packets ({:.1%}) – last from {}".format(nprocessed, npackets,
                                                                                     float(nprocessed) / npackets,
                                                                                     worker))
            else:
                logger.info(
                    "Processed {}/{} packets ({:.1%})".format(nprocessed, npackets, float(nprocessed) / npackets))
            logger.debug("Mem usage: processed packet #{}: {}".format(nprocessed, proc.memory_info().rss))

        logger.debug("Mem usage: results processed: {}".format(proc.memory_info().rss))

        for uid, message in global_warnings.iteritems():
            logger.warn("UID {}: {}".format(uid, message))

        with open(opts.bin_list, 'w') as f_bin_list:
            global_widths['uid'] -= 2
            f_bin_list.write("# {:>{uid}} {:>{binpath}} {:>{file_size}} {:>{seqlen}} "
                             "{:>{npoints}} {:>{nuniq}} {:>{window_size}} {:>{npoints_per_res}} "
                             "{:>{black_score}} {:>{log_score}}\n".format(*saved_fields, **global_widths))
            global_widths['uid'] += 2
            for arrays in saved_results:
                for r in izip(*arrays):
                    uid, status, existence, file_size, seqlen, npoints, nuniq, window_size, npoints_per_res, black_score, log_score = r
                    binpath = create_binpath(uid, opts.depth, opts.output_format)
                    if status == 0:
                        if not opts.print_only:
                            if file_size == 0:
                                logger.warning(
                                    "!!! File size of {} is {}, job status is {}".format(binpath, file_size, status))
                                continue
                            kept_file_size += file_size
                            kept_files += 1
                            total_file_size += file_size

                        f_bin_list.write("{:>{uid}} {:>{binpath}} {:>{file_size}} {:>{seqlen}} "
                                         "{:>{npoints}} {:>{nuniq}} {:>{window_size}} {:>{npoints_per_res}.3f} "
                                         "{:>{black_score}.3f} {:>{log_score}.3f}\n".format(uid, binpath, file_size,
                                                                                            seqlen,
                                                                                            npoints, nuniq, window_size,
                                                                                            npoints_per_res,
                                                                                            black_score,
                                                                                            log_score, **global_widths))
                    else:
                        logger.warning("Sequence {} was not processed properly (status {})".format(uid, status))
        logger.info(
            "Processed {} sequences, kept {}. Created {:.3e} bytes, "
            "kept files occupy {:.3e} bytes".format(nsequences, kept_files, total_file_size, kept_file_size))

        logger.debug("Mem usage: results written: {}".format(proc.memory_info().rss))
    else:
        logger.warn("No sequences fulfilling criteria to run dotter.")
        if opts.filter:
            logger.warn("Perhaps 'UniRef90_' prefix is missing on the list?")


if __name__ == "__main__":
    main()
