#!/usr/bin/python2
# coding=utf-8
import csv
import time
import logging
from argparse import ArgumentParser, FileType

from Bio import SeqIO
import os


def main():
    parser = ArgumentParser(description="Put multi-sequence fasta file into one-sequence-per-line csv")
    parser.add_argument("-f", "--fasta", type=FileType('r'), dest="fasta", required=True,
                        help="Fasta datafile with sequence id and sequence itself")
    parser.add_argument("-o", "--csv-out", type=FileType('w'), dest="csv_out", required=True,
                        help="CSV file with one line per sequence, good for further splitting "
                             "when the data set is large")
    parser.add_argument("-p", "--progress-interval", dest="progress_interval", default=120, type=float,
                        help="Time interval for printing progress (seconds)")
    parser.add_argument("-l", "--logfile", dest="logfile", default="fasta2csv.log", help="Log file name")

    opts = parser.parse_args()

    if os.path.exists(opts.logfile):
        i = 1
        while os.path.exists("{}.{}".format(opts.logfile, i)):
            i += 1
        os.rename(opts.logfile, "{}.{}".format(opts.logfile, i))

    logger = logging.getLogger("fasta2csv")
    logger.setLevel(logging.INFO)
    fh = logging.FileHandler(opts.logfile, 'w')
    ch = logging.StreamHandler()
    ch.setLevel(logging.ERROR)
    # create formatter and add it to the handlers
    formatter = logging.Formatter('[%(asctime)-15s] %(name)-9s %(levelname)-7s %(message)s')
    fh.setFormatter(formatter)
    ch.setFormatter(formatter)
    # add the handlers to the logger
    logger.addHandler(fh)
    logger.addHandler(ch)
    logger.info("Processing started")
    csv_writer = csv.writer(opts.csv_out, delimiter=';')
    csv_writer.writerow(["ID", "length", "sequence"])
    sequences_written = 0
    start = time.time()
    for fdata in SeqIO.FastaIO.FastaIterator(opts.fasta):
        seq = str(fdata.seq)
        csv_writer.writerow([fdata.id, len(seq), seq])
        sequences_written += 1
        current = time.time()
        if current - start > opts.progress_interval:
            logger.info("Processed {} sequences".format(sequences_written))
            start += opts.progress_interval

    logger.info("FINISHED: wrote {} sequences".format(sequences_written))


if __name__ == "__main__":
    main()
