#include <exception>
#include <vector>
#include <cstring>
#include <algorithm>
#include <sstream>
#include <sys/stat.h>
#include <sys/sendfile.h>
#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>
#include <omp.h>
#include <chrono>
#include "DotPlotCollection.h"

unsigned DotPlotCollection::_nDotPlots = 0;
unsigned long DotPlotCollection::_nResults = 0;
off64_t DotPlotCollection::_contentsPosition = 0;
ushort DotPlotCollection::_maxSeqLen[2] = {0, 0};
ushort DotPlotCollection::_maxNUniq = 0;
std::string DotPlotCollection::_dataFileName = "";
char *DotPlotCollection::_outFilePrefix = nullptr;
FILE *DotPlotCollection::_fpDataFile = nullptr;
int DotPlotCollection::_fdescDataFile = -1;
int DotPlotCollection::_nThreads = 1;
const unsigned char DotPlotCollection::RECORD_START = 187; // 0xBB
const unsigned int DotPlotCollection::DP_VERSION = 4;
const char *DotPlotCollection::_header = "Dot plot stash";

void DotPlotCollection::readCollectionHeader() {

    // Open data file for reading
    _fpDataFile = fopen(_dataFileName.c_str(), "rb");
    if (_fpDataFile == NULL) {
        perror("Could not open input file");
        throw std::runtime_error("I/O Error");
    }
    // allocate
    char headerIn[16];
    uint fversion = 0;

    if (fread(headerIn, sizeof(char), strlen(_header), _fpDataFile) != strlen(_header)) {
        fprintf(stderr, "Header size different than expected");
        throw std::runtime_error("Unmatched header size");
    }
    printf("%s\n", headerIn);
    if (strcmp(_header, headerIn) != 0) {
        throw std::runtime_error("The header strings are not equal! Possible corruption");
    }
    if (fread(&fversion, sizeof(uint), 1, _fpDataFile) != 1) {
        throw std::runtime_error("Incorrect file version size");
    }
    if (fversion != DP_VERSION) {
        fprintf(stderr, "The current code version %d does not match the file version %d\n", DP_VERSION, fversion);
        throw std::runtime_error("Unmatched code/input file version");
    }
    if (fread(&_nDotPlots, sizeof(uint), 1, _fpDataFile) != 1) {
        throw std::runtime_error("Could not read the number of dot plots properly");
    }

    /* Important: overflow-safe */
    _nResults = _nDotPlots;
    _nResults *= _nDotPlots;

    fprintf(stdout, "The input file contains %u dot plots\n", _nDotPlots);

    _contentsPosition = ftello64(_fpDataFile);
}

void DotPlotCollection::writeCollectionHeader() {
    if (_fpDataFile)
        throw std::runtime_error("This file was open before");

    _fpDataFile = fopen64(_dataFileName.c_str(), "wb");
    if (_fpDataFile == nullptr) {
        perror("Could not open output file");
        throw std::runtime_error("I/O Error");
    }
    fwrite(_header, sizeof(char), strlen(_header), _fpDataFile);

    fwrite(&DP_VERSION, sizeof(uint), 1, _fpDataFile);

    fwrite(&_nDotPlots, sizeof(unsigned), 1, _fpDataFile);

    _contentsPosition = ftello64(_fpDataFile);
}

void DotPlotCollection::prepareComparisonMatrix() {
    off64_t sizeOutFile = _nResults * sizeof(float);
    std::string outFileNameStr = getComparisonMatrixFileName();

    fprintf(stdout, "--- Creating output file %s of size %li\n", outFileNameStr.c_str(), sizeOutFile);

    FILE *_fpOutFile = fopen(outFileNameStr.c_str(), "wb+");
    if (!_fpOutFile) {
        perror("The following i/o error occured:");
    }
    if (fseeko64(_fpOutFile, sizeOutFile - 1, SEEK_SET) == -1) {
        perror("The following i/o error occured:");
    }
    char zero = 0;
    fwrite(&zero, sizeof(char), 1, _fpOutFile);
    if (ftello64(_fpOutFile) != sizeOutFile) {
        fprintf(stderr, "Could not set file size properly. Got %li instead %li\n", ftello64(_fpOutFile), sizeOutFile);
        throw std::runtime_error("I/O Error");
    }
    fclose(_fpOutFile);
}

void DotPlotCollection::checkComparisonMatrix() {
    struct stat buffer;
    std::string outFileNameStr = getComparisonMatrixFileName();
    if (stat(outFileNameStr.c_str(), &buffer) != 0)
        throw std::runtime_error("Output file not found although it should exist at this point");
}

void DotPlotCollection::populate() {

    fprintf(stdout, "Reading data set with indices <%d, %d>\n", _dataRange->getStart(),
            _dataRange->getStart() + _dataRange->getRange() - 1);
    fflush(stdout);
    _contents = new off64_t[_dataRange->getRange()];
    _data = new DotPlot[_dataRange->getRange()];

    // find a starting position in the contents list
    if (fseeko64(_fpDataFile, _contentsPosition + _dataRange->getStart() * sizeof(off64_t), SEEK_SET) != 0) {
        perror("The following error occured:");
    }

    fprintf(stdout, "We begin at pos %li\n", ftello64(_fpDataFile));
    // read list of pointers
    size_t ncread = fread(_contents, sizeof(off64_t), _dataRange->getRange(), _fpDataFile);
    if (ncread != _dataRange->getRange()) {
        fprintf(stderr, "!!! Did not read as many elements as expected (%zu vs %d)\n", ncread, _dataRange->getRange());
        throw std::runtime_error("!!! Contents reading error");
    }

    // seek to the beginning of data
    fseeko64(_fpDataFile, _contents[0], SEEK_SET);

    DotPlot *pr;
    unsigned char section;
    size_t nread, nZeroPoints = 0;

    for (uint pos = 0; pos < _dataRange->getRange(); ++pos) {
        pr = &_data[pos];
        if (ftello64(_fpDataFile) != _contents[pos]) {
            fprintf(stderr, "Dot plot %u: Real position: %li, precalculated: %li\n",
                    pos, ftello64(_fpDataFile), _contents[pos]);
            throw std::runtime_error("!!! Real file position does not match the precalculated one");
        }

        if (fread(&section, sizeof(u_char), 1, _fpDataFile) != 1) {
            fprintf(stderr, "Could not read section marker\n");
            throw std::runtime_error("!!! Input file read error");
        }

        if (section != RECORD_START) {
            fprintf(stderr, "File possibly corrupted!");
            throw std::runtime_error("!!! Input file read error");
        }

        if (fread(&pr->nameSize, sizeof(ushort), 1, _fpDataFile) != 1) {
            fprintf(stderr, "Could not read sequence id length\n");
            throw std::runtime_error("!!! Input file read error");
        }

        pr->name = new u_char[pr->nameSize + 1];
        pr->name[pr->nameSize] = static_cast<u_char>('\0');
        nread = fread(pr->name, sizeof(u_char), pr->nameSize, _fpDataFile);
        if (nread * sizeof(u_char) != pr->nameSize) {
            fprintf(stderr, "String sizes are not in agreement for %d\n", pos);
            throw std::runtime_error("!!! Input file read error");
        }

        if (fread(&pr->seqLen, sizeof(ushort), 1, _fpDataFile) != 1) {
            fprintf(stderr, "Could not read sequence length\n");
            throw std::runtime_error("!!! Input file read error");
        }

        if (fread(&pr->windowSize, sizeof(ushort), 1, _fpDataFile) != 1) {
            fprintf(stderr, "Could not read window size\n");
            throw std::runtime_error("!!! Input file read error");
        }

        if (fread(&pr->nPoints, sizeof(uint), 1, _fpDataFile) != 1) {
            fprintf(stderr, "Could not read the number of points\n");
            throw std::runtime_error("!!! Input file read error");
        }
        if (pr->nPoints == 0)
            ++nZeroPoints;

        if (fread(&pr->nPointGroups, sizeof(uint), 1, _fpDataFile) != 1) {
            fprintf(stderr, "Could not read the number of point groups\n");
            throw std::runtime_error("!!! Input file read error");
        }

        if (fread(&pr->nUniq, sizeof(ushort), 1, _fpDataFile) != 1) {
            fprintf(stderr, "Could not read the number of unique offsets\n");
            throw std::runtime_error("!!! Input file read error");
        }

        pr->uniqueOffsets = new ushort[pr->nUniq];
        pr->nPointGroupsPerUniqueOffset = new ushort[pr->nUniq];
        pr->cumulativeNGroupsPerUniqueOffset = new uint[pr->nUniq];
        pr->pointGroupStart = new ushort[pr->nPointGroups];
        pr->pointGroupSize = new ushort[pr->nPointGroups];

        if (fread(pr->uniqueOffsets, sizeof(ushort), pr->nUniq, _fpDataFile) != pr->nUniq) {
            fprintf(stderr, "Could not read the unique offsets\n");
            throw std::runtime_error("!!! Input file read error");
        }

        if (fread(pr->nPointGroupsPerUniqueOffset, sizeof(ushort), pr->nUniq, _fpDataFile) != pr->nUniq) {
            fprintf(stderr, "Could not read the number of groups per unique offset\n");
            throw std::runtime_error("!!! Input file read error");
        }

        if (fread(pr->cumulativeNGroupsPerUniqueOffset, sizeof(uint), pr->nUniq, _fpDataFile) != pr->nUniq) {
            fprintf(stderr, "Could not read the cumulative number of point groups\n");
            throw std::runtime_error("!!! Input file read error");
        }

        if (fread(pr->pointGroupStart, sizeof(ushort), pr->nPointGroups, _fpDataFile) !=
            pr->nPointGroups) {
            fprintf(stderr, "Could not read the point group start data\n");
            throw std::runtime_error("!!! Input file read error");
        }

        if (fread(pr->pointGroupSize, sizeof(ushort), pr->nPointGroups, _fpDataFile) !=
            pr->nPointGroups) {
            fprintf(stderr, "Could not read the point group size data\n");
            throw std::runtime_error("!!! Input file read error");
        }

        // Calculate two maximum sequence lengths
        if (pr->seqLen > _maxSeqLen[sequenceLength::first]) {
            // move 1st to 2nd position
            _maxSeqLen[sequenceLength::second] = _maxSeqLen[sequenceLength::first];
            _maxSeqLen[sequenceLength::first] = pr->seqLen;
        } else if (pr->seqLen > _maxSeqLen[sequenceLength::second])
            _maxSeqLen[sequenceLength::second] = pr->seqLen;

        // Calculate maximum number of unique points
        if (pr->nUniq > _maxNUniq)
            _maxNUniq = pr->nUniq;
    }

    fprintf(stdout, ">>> Maximum sequence lengths are %d and %d\n",
            _maxSeqLen[sequenceLength::first], _maxSeqLen[sequenceLength::second]);
    fprintf(stdout, ">>> Maximum number of unique points is %d\n", _maxNUniq);
    if (nZeroPoints)
        fprintf(stdout, "!! Found %zu sequences with no points\n", nZeroPoints);
    else
        fprintf(stdout, "^^ All sequences contain points\n");

    fflush(stdout);
}

void DotPlotCollection::compare(const DotPlotCollection &rhs) {
    auto start = std::chrono::steady_clock::now();

    // close input
    fclose(_fpDataFile);

    // Maximum number of possible slides
    const int nSlideMax = _maxSeqLen[sequenceLength::first] + _maxSeqLen[sequenceLength::second] - 2;

    // DONE result buffer
    int_fast64_t nLocalResultBufferSize = _dataRange[0] ^_dataRange[1];
    float *results = new float[nLocalResultBufferSize];
    for (size_t k = 0; k < nLocalResultBufferSize; k++)
        results[k] = 0.0f;

    auto stop = std::chrono::steady_clock::now();
    auto duration = std::chrono::duration_cast<std::chrono::seconds>(stop - start);
    fprintf(stdout, "---- Subtimer [%li s] Started calculations\n", duration.count());
    fflush(stdout);

#pragma omp parallel default(none) shared(rhs, results, stderr)
    {
        std::string outFileNameStr = getComparisonMatrixFileName();
        uint *slideCollection = new uint[nSlideMax];
        for (size_t k = 0; k < nSlideMax; k++) {
            slideCollection[k] = 0;
        }

        /* MAIN LOOP
         * Only external loop is parallelized
         * Dynamic schedule used since task load will differ
         * Explicit stating which variable is shared and which is private for better readability and bug avoidance
         * */
        omp_set_nested(1);
#pragma omp for schedule(dynamic)
        for (int i = 0; i < _dataRange->getRange(); i++) {
            /* Initialize */
            int startJ = (_reference || rhs._reference) ? i + 1 : 0;
            /* Important to avoid overflow */
            int_fast64_t index = i;
            index *= rhs._dataRange->getRange();
            index += startJ;

            if (_data[i].nPoints == 0) {
                for (int j = startJ; j < rhs._dataRange->getRange(); j++) {
                    results[index++] = 0;
                }
                continue;
            }

            const DotPlot *ps, *pl;

            for (int j = startJ; j < rhs._dataRange->getRange(); j++) {
                if (rhs._data[j].nPoints == 0) {
                    results[index++] = 0;
                    continue;
                }

                if (_data[i].seqLen > rhs._data[j].seqLen) {
                    pl = &_data[i];
                    ps = &rhs._data[j];
                } else {
                    pl = &rhs._data[j];
                    ps = &_data[i];
                }

                // For a given pair, we only need this small padding
                uint padding = static_cast<uint>(ps->seqLen - 1);
                int nSlidePairMax = ps->seqLen + pl->seqLen - 2;

                const ushort *ulgStart = pl->uniqueOffsets;
                const ushort *ulgEnd = pl->uniqueOffsets + pl->nUniq;
                const ushort *usmStart = ps->uniqueOffsets;
                const ushort *usmEnd = ps->uniqueOffsets + ps->nUniq;

                ushort lci = 0, sci = 0;

                while (ulgStart != ulgEnd && usmStart != usmEnd) {
                    if (*ulgStart < *usmStart) {
                        ++ulgStart;
                        ++lci;
                    } else if (*ulgStart > *usmStart) {
                        ++usmStart;
                        ++sci;
                    } else {
                        // NEW START

                        // calculate Jaccard index
                        uint innerStart =
                                ps->cumulativeNGroupsPerUniqueOffset[sci] - ps->nPointGroupsPerUniqueOffset[sci];
                        for (uint kl = pl->cumulativeNGroupsPerUniqueOffset[lci] - pl->nPointGroupsPerUniqueOffset[lci];
                             kl < pl->cumulativeNGroupsPerUniqueOffset[lci];
                             ++kl) {
                            ushort largeDotPlotGroupStart = pl->pointGroupStart[kl];
                            ushort largeDotPlotGroupLength = pl->pointGroupSize[kl];

                            for (uint ks = innerStart;
                                 ks < ps->cumulativeNGroupsPerUniqueOffset[sci];
                                 ++ks) {

                                // this refers to group size of small dot plot
                                ushort smallDotPlotGroupStart = ps->pointGroupStart[ks];
                                ushort smallDotPlotGroupLength = ps->pointGroupSize[ks];
                                // this refers to the size of a group regardless of the dot plot size
                                const ushort *smallGroupLength, *largeGroupLength;
                                if (largeDotPlotGroupLength >= smallDotPlotGroupLength) {
                                    largeGroupLength = &largeDotPlotGroupLength;
                                    smallGroupLength = &smallDotPlotGroupLength;
                                } else {
                                    largeGroupLength = &smallDotPlotGroupLength;
                                    smallGroupLength = &largeDotPlotGroupLength;
                                }
                                // if groups are large
                                if (*smallGroupLength > 2 && *largeGroupLength > 5) {
                                    uint counts = 0;
                                    uint paddedAMinB = padding + largeDotPlotGroupStart - smallDotPlotGroupStart;
                                    uint top = paddedAMinB + largeDotPlotGroupLength - 1;
                                    uint bottom = paddedAMinB - smallDotPlotGroupLength + 1;

                                    while (++counts < *smallGroupLength) {
                                        slideCollection[bottom++] += counts;
                                        slideCollection[top--] += counts;
                                    }

                                    while (counts++ <= *largeGroupLength) {
                                        slideCollection[bottom++] += *smallGroupLength;
                                    }
                                } else {
                                    // if they are small, do it traditionally
                                    uint il = largeDotPlotGroupStart;
                                    while (il < largeDotPlotGroupStart + largeDotPlotGroupLength) {
                                        uint is = smallDotPlotGroupStart;
                                        uint paddedStartIndex = padding + il - is;
                                        while (is++ < smallDotPlotGroupStart + smallDotPlotGroupLength) {
                                            ++slideCollection[paddedStartIndex--];
                                        }
                                        ++il;
                                    }
                                }
                            }
                        }

                        // NEW END

                        ++lci;
                        ++sci;

                        ++ulgStart;
                        ++usmStart;
                    }
                } // END while

                uint A_intersection_B = 0;

                for (int k = 0; k < nSlidePairMax; ++k) {
                    if (slideCollection[k] > A_intersection_B)
                        A_intersection_B = slideCollection[k];

                    // Clean the slide collection for this thread if it was used
                    slideCollection[k] = 0;
                }

                //results[index] = 0;
                if (A_intersection_B > 0)
                    results[index] =
                            static_cast<float>(A_intersection_B) /
                            (pl->nPoints + ps->nPoints - A_intersection_B);

                ++index;
            }
        } // END omp for

        delete[] slideCollection;

    } // END omp parallel section

    stop = std::chrono::steady_clock::now();
    duration = std::chrono::duration_cast<std::chrono::seconds>(stop - start);
    fprintf(stdout, "---- Subtimer [%li s] Finished calculations\n", duration.count());
    fflush(stdout);

    std::string outFileNameStr = getComparisonMatrixFileName();

    _fdescDataFile = open(outFileNameStr.c_str(), O_RDWR);
    if (_fdescDataFile < 0) {
        perror("Could not open results file:");
        throw std::runtime_error("I/O Error");
    }

    stop = std::chrono::steady_clock::now();
    duration = std::chrono::duration_cast<std::chrono::seconds>(stop - start);
    fprintf(stdout, "---- Subtimer [%li s] Opened output file\n", duration.count());
    fflush(stdout);

    if (_reference || rhs._reference) {
        // self-self comparison, i.e. zeroes on diagonal
        const unsigned nRecords = _dataRange->getRange();
        for (size_t iRow = 0; iRow < nRecords; iRow++)
            for (size_t iCol = iRow + 1; iCol < nRecords; iCol++) {
                results[iCol * nRecords + iRow] = results[iRow * nRecords + iCol];
            }

        stop = std::chrono::steady_clock::now();
        duration = std::chrono::duration_cast<std::chrono::seconds>(stop - start);
        fprintf(stdout, "---- Subtimer [%li s] Filled lower triangle\n", duration.count());
        fflush(stdout);

        writeResults(results, _dataRange->getStart(), rhs._dataRange->getStart(), nRecords, nRecords);
        stop = std::chrono::steady_clock::now();
        duration = std::chrono::duration_cast<std::chrono::seconds>(stop - start);
        fprintf(stdout, "---- Subtimer [%li s] Written diagonal chunk\n", duration.count());
        fflush(stdout);

    } else {
        // write to file
        writeResults(results, _dataRange->getStart(), rhs._dataRange->getStart(),
                     _dataRange->getRange(), rhs._dataRange->getRange());
        stop = std::chrono::steady_clock::now();
        duration = std::chrono::duration_cast<std::chrono::seconds>(stop - start);
        fprintf(stdout, "---- Subtimer [%li s] Written upper-diagonal chunk\n", duration.count());
        fflush(stdout);

        // square matrix
        if (_dataRange->getRange() == rhs._dataRange->getRange()) {
            float tmp;
            const unsigned nRecords = _dataRange->getRange();
            size_t currentPosition = 0, newPosition = 0;
            // simple transposition
            fprintf(stdout, "Performing a simple transpose of a square chunk\n");
            fflush(stdout);
#pragma omp parallel for schedule(dynamic) default(none) private(currentPosition, newPosition, tmp) shared(results)
            for (size_t iRow = 0; iRow < nRecords; iRow++) {
                currentPosition = iRow * nRecords + iRow + 1;
                newPosition = (iRow + 1) * nRecords + iRow;
                for (size_t iCol = iRow + 1; iCol < nRecords; iCol++) {
                    tmp = results[newPosition];
                    results[newPosition] = results[currentPosition];
                    results[currentPosition] = tmp;
                    ++currentPosition;
                    newPosition += nRecords;
                }
            }
        } else {
            unsigned nNewRows = rhs._dataRange->getRange();
            unsigned nNewCols = _dataRange->getRange();
            int_fast64_t tmpNLocalResults = nNewRows;
            tmpNLocalResults *= nNewCols;
            const int_fast64_t nLocalResults = tmpNLocalResults;

            float *first = results;
            float *last = results + nLocalResults;
            const int_fast64_t size_min_1 = nLocalResults - 1;

            std::vector<bool> visited(nLocalResults);

            fprintf(stdout, "Performing a complex transpose of a non-square chunk\n");
            fflush(stdout);

            float *cycle = first;
            while (++cycle != last) {
                if (visited[cycle - first])
                    continue;
                unsigned long currentIndex = cycle - first;
                do {
                    currentIndex = currentIndex == size_min_1 ? size_min_1 : (nNewCols * currentIndex) % size_min_1;
                    std::swap(*(first + currentIndex), *cycle);
                    visited[currentIndex] = true;
                } while ((first + currentIndex) != cycle);
            }
        }
        stop = std::chrono::steady_clock::now();
        duration = std::chrono::duration_cast<std::chrono::seconds>(stop - start);
        fprintf(stdout, "---- Subtimer [%li s] Finished chunk transposition\n", duration.count());
        fflush(stdout);

        // write to file in a new position
        writeResults(results, rhs._dataRange->getStart(), _dataRange->getStart(),
                     rhs._dataRange->getRange(), _dataRange->getRange());

        stop = std::chrono::steady_clock::now();
        duration = std::chrono::duration_cast<std::chrono::seconds>(stop - start);
        fprintf(stdout, "---- Subtimer [%li s] Written transposed chunk\n", duration.count());
        fflush(stdout);
    }

    close(_fdescDataFile);

    delete[] results;
}


void
DotPlotCollection::writeResults(const float *results, const unsigned &rowStart, const unsigned &colStart,
                                const unsigned &nRows,
                                const unsigned &nCols) {
// write EVERYTHING to file
    ssize_t written = 0;
    size_t nAll = nRows;
    nAll *= nCols;

    // save from integer overflow
    off64_t offsetStart = rowStart;
    offsetStart *= _nDotPlots;
    offsetStart += colStart;
    offsetStart *= sizeof(float);
    off64_t offsetShift = (_nDotPlots - nCols) * sizeof(float);
    off64_t lineWidth = _nDotPlots * sizeof(float);

    auto start = std::chrono::steady_clock::now();

    ssize_t expectedCount = sizeof(float) * nCols;
    // Satisfy OpenMP specs
    int fdescDataFileCopy = _fdescDataFile;
#pragma omp parallel for schedule(dynamic) default(none) \
        shared(expectedCount, offsetStart, lineWidth, results, nCols, nRows, stderr, fdescDataFileCopy) private(written)
    for (size_t k = 0; k < nRows; k++) {
        if ((written = pwrite(fdescDataFileCopy, results + k * nCols, expectedCount, offsetStart + k * lineWidth)) !=
            expectedCount) {
            if (written >= 0) {
                fprintf(stderr, "Wrote %zd values instead of %zd\n", written, expectedCount);
            } else {
                perror("Could not write the buffer to a file:\n");
            }
            throw std::runtime_error("I/O Error");
        }
    }

    auto stop = std::chrono::steady_clock::now();
    auto duration = std::chrono::duration_cast<std::chrono::seconds>(stop - start);

    fprintf(stdout,
            ">>> Wrote %zu elements in %d rows, %d columns each, "
            "from (%d, %d) – start %lu, jump %lu\n",
            nAll, nRows, nCols, rowStart, colStart, offsetStart, offsetShift);
    // Compute write rate
    long writeTime = duration.count();
    if (writeTime > 0) {
        double writeRate = sizeof(float) * double(nAll) / writeTime;
        double wRLog = std::log10(writeRate);
        const char *prefix[] = {"", "k", "M", "G"};
        int prefixIndex = int(wRLog) / 3;
        if (prefixIndex > 3)
            prefixIndex = 3;
        int wRDivider = prefixIndex * 3;
        if (wRDivider > 0)
            writeRate /= std::pow(10, wRDivider);
        fprintf(stdout, "---- Write rate %5.3f %sB/s\n", writeRate, prefix[prefixIndex]);
    } else {
        fprintf(stdout, "---- Write rate: inf B/s\n");
    }

    fflush(stdout);
}

DotPlotCollection::~DotPlotCollection() {

    if (!_reference) {
        delete[] _data;
        delete[] _contents;
    }
}

int_fast64_t DotPlotCollection::nTotalJobs() {
    int_fast64_t result = _nDotPlots;
    result *= (_nDotPlots - 1);
    return result / 2;
}

DotPlotCollection &DotPlotCollection::operator=(const DotPlotCollection &rhs) {
    _dataRange = rhs._dataRange;
    _contents = rhs._contents;
    _data = rhs._data;
    _reference = true;
    return *this;
}

std::string DotPlotCollection::getComparisonMatrixFileName() {
    return Utils::generateFileName(_outFilePrefix, "results", "bin");
}

void DotPlotCollection::buildFromInfoRepository(DotPlotInfoRepository &dpInfoRepository) {
    _nDotPlots = (unsigned int) dpInfoRepository.getNDotPlots();
    writeCollectionHeader();

    _contents = new off64_t[_nDotPlots];
    _contents[0] = _contentsPosition + sizeof(off64_t) * _nDotPlots;
    for (size_t i = 1; i < _nDotPlots; i++) {
        _contents[i] = _contents[i - 1] + dpInfoRepository[i - 1].fileSize + 1; // 1 for RECORD_START (char)
    }
    fwrite(_contents, sizeof(off64_t), _nDotPlots, _fpDataFile);
    // inflate the file to the full size
    fseeko64(_fpDataFile, dpInfoRepository.getRawDataSize() + _nDotPlots - 1, SEEK_CUR);
    char zero = 0;
    fwrite(&zero, sizeof(char), 1, _fpDataFile);
    fclose(_fpDataFile);

    // distribute data writing between threads
    size_t *dotPlotsPerThread = new size_t[_nThreads * 2];
    size_t meanDataSizePerThread = dpInfoRepository.getRawDataSize() / _nThreads;
    size_t j = 0;
    int k = 0;
    dotPlotsPerThread[k] = 0;
    for (int i = 0; i < _nThreads; i++) {
        size_t perThread = 0;
        while (perThread < meanDataSizePerThread && j < _nDotPlots) {
            perThread += dpInfoRepository[j++].fileSize;
        }
        if ((perThread - meanDataSizePerThread) / dpInfoRepository[j - 1].fileSize > 0.5) {
            perThread -= dpInfoRepository[j - 1].fileSize;
            --j;
        }
        if (i == _nThreads - 1)
            while (j < _nDotPlots)
                perThread += dpInfoRepository[j++].fileSize;
        k = 2 * i;
        if (k > 1)
            dotPlotsPerThread[k] = dotPlotsPerThread[k - 2] + dotPlotsPerThread[k - 1];
        dotPlotsPerThread[k + 1] = j - dotPlotsPerThread[k];
        fprintf(stdout, "Thread %d will start from %zu and process %zu dot plots, total %zu bytes\n",
                i, dotPlotsPerThread[k], dotPlotsPerThread[k + 1], perThread);
    }

    // these two lines were enforced by OpenMP specs
    off64_t *contentsPtr = _contents;
    std::string dataFileNameRef = _dataFileName;
    //
#pragma omp parallel default(none) shared(dataFileNameRef, contentsPtr, dpInfoRepository, dotPlotsPerThread, stderr)
    {
        int foutDataThread = open(dataFileNameRef.c_str(), O_WRONLY);
        if (foutDataThread == -1) {
            fprintf(stderr, "Could not open output %s\nDetails: %s\n", dataFileNameRef.c_str(), strerror(errno));
            throw std::runtime_error("I/O Error");
        }
        char buffer[BUFSIZ];
        ssize_t bytesRead;

        int threadNum = omp_get_thread_num();
        int n = threadNum * 2;
        ssize_t bytesWritten;
        off64_t currentContents = contentsPtr[dotPlotsPerThread[n]];
        for (size_t m = dotPlotsPerThread[n]; m < dotPlotsPerThread[n] + dotPlotsPerThread[n + 1]; m++) {
            ssize_t totalBytesWritten = 0;
            int fIndividualDataSet = open(dpInfoRepository[m].path.c_str(), O_RDONLY);
            if (fIndividualDataSet == -1) {
                fprintf(stderr, "Could not open input %s\nDetails: %s\n", dpInfoRepository[m].path.c_str(),
                        strerror(errno));
                throw std::runtime_error("I/O Error");
            }

            bytesWritten = pwrite(foutDataThread, &RECORD_START, sizeof(u_char), currentContents);
            currentContents += bytesWritten;
            totalBytesWritten += bytesWritten;
            while ((bytesRead = read(fIndividualDataSet, (char *) buffer, BUFSIZ)) > 0) {
                bytesWritten = pwrite(foutDataThread, (char *) buffer, static_cast<size_t>(bytesRead), currentContents);
                if (bytesRead != bytesWritten) {
                    fprintf(stderr, "Error copying contents of input file %s\n", dpInfoRepository[m].path.c_str());
                    throw std::runtime_error("I/O Error");
                }
                currentContents += bytesWritten;
                totalBytesWritten += bytesWritten;
            }

            if (totalBytesWritten != dpInfoRepository[m].fileSize + sizeof(u_char)) {
                fprintf(stderr, "Wrote %zi bytes instead of %zu for %s\n", totalBytesWritten,
                        dpInfoRepository[m].fileSize + sizeof(u_char), dpInfoRepository[m].path.c_str());
                throw std::runtime_error("I/O Error");
            }

            close(fIndividualDataSet);
        }
        close(foutDataThread);
    }

    delete[] dotPlotsPerThread;
}
