#include <cstring>
#include <stdexcept>
#include "DotPlotInfoRepository.h"

DotPlotInfoRepository::DotPlotInfoRepository(const char *infoFileName,
                                             const DotPlotInfoItemIndex &lastColumnToRead) {
    nDotPlots = 0;
    totalABCSize = 0;
    rawDataSize = 0;
    repoFileName = infoFileName;
    lastColumn = lastColumnToRead;
}

void DotPlotInfoRepository::process() {
    if (repoFileName == nullptr) {
        throw std::runtime_error("File name for dot plot info repository is null");
    }

    fprintf(stdout, "Reading protein sequence data %s\n", repoFileName);
    char *line = nullptr, *token = nullptr, *cpline = nullptr;
    unsigned long long sIDSum = 0;
    size_t len = 0, nDataLines = 0;
    int nTokensPerCurrentLine;
    const char *delimiters = " \t";

    DotPlotInfoItem tmp;
    FILE *fpInfo = fopen64(repoFileName, "r");
    if (fpInfo == nullptr) {
        perror("Could not read input file:");
        throw std::runtime_error("Reading dot plot info file failed");
    }

    while (getline(&line, &len, fpInfo) != -1) {
        cpline = line;
        token = strtok(line, delimiters);
        nTokensPerCurrentLine = 0;

        // ignore comment
        if (line[0] == '#')
            continue;
        else {
            tmp.id = token;
            totalABCSize += nDataLines * strlen(token) + sIDSum;
            sIDSum += strlen(token);
            ++nTokensPerCurrentLine;
        }

        while ((token = strtok(nullptr, delimiters)) != nullptr || nTokensPerCurrentLine <= lastColumn) {
            switch (nTokensPerCurrentLine) {
                case DotPlotInfoItemIndex::DP_PATH:
                    tmp.path = token;
                    break;
                case DotPlotInfoItemIndex::DP_FS:
                    tmp.fileSize = static_cast<size_t>(std::stoul(token));
                    break;
                case DotPlotInfoItemIndex::DP_SEQL:
                    tmp.seqLen = static_cast<unsigned short>(std::stoi(token));
                    break;
                case DotPlotInfoItemIndex::DP_NPOINTS:
                    tmp.nPoints = static_cast<unsigned int>(std::stoi(token));
                    break;
                case DotPlotInfoItemIndex::DP_NUNIQ:
                    tmp.nUniq = static_cast<unsigned short>(std::stoi(token));
                    break;
                case DotPlotInfoItemIndex::DP_WINSIZE:
                    tmp.windowSize = static_cast<unsigned short>(std::stoi(token));
                    break;
                case DotPlotInfoItemIndex::DP_NPOINTSPERRES:
                    tmp.nPointsPerResidue = std::stof(token);
                    break;
                case DotPlotInfoItemIndex::DP_BLACKSCORE:
                    tmp.blackScore = std::stof(token);
                    break;
                case DotPlotInfoItemIndex::DP_LOGSCORE:
                    tmp.logScore = std::stof(token);
                    break;
                default:
                    fprintf(stderr, "!!! Found %d tokens\n", nTokensPerCurrentLine);
                    throw std::runtime_error("Too many tokens per line");
            }
            ++nTokensPerCurrentLine;
        }

        if (nTokensPerCurrentLine != DotPlotInfoItemIndex::DP_NR) {
            fprintf(stderr, "!!! This line\n%s\n has %d tokens while expected %d\n",
                    cpline, nTokensPerCurrentLine, DotPlotInfoItemIndex::DP_NR);
            throw std::runtime_error("All data lines should have equal number of tokens");
        }

        dotPlotInfoCollection.emplace_back(tmp);
        rawDataSize += tmp.fileSize;
        ++nDataLines;
    }

    fclose(fpInfo);

    if (line)
        free(line);

    /* The number of proteins in the set */
    nDotPlots = dotPlotInfoCollection.size();

    if (nDataLines != nDotPlots)
        fprintf(stderr, "!!! Number of data lines %zu is not equal to the number of dot plots %zu\n",
                nDataLines, nDotPlots);
    fprintf(stdout, ">>> Found %zu dotplots of total size %zu bytes\n", nDataLines, rawDataSize);
}

DotPlotInfoItem &DotPlotInfoRepository::operator[](size_t index) {
    return dotPlotInfoCollection[index];
}

const size_t &DotPlotInfoRepository::getNDotPlots() const {
    return nDotPlots;
}

const size_t &DotPlotInfoRepository::getRawDataSize() const {
    return rawDataSize;
}

