#include <iostream>
#include <getopt.h>
#include <cstring>
#include <chrono>
#include <omp.h>
#include "utils.h"
#include "DotPlot.h"
#include "DataRange.h"
#include "DotPlotCollection.h"

static void usage() {
    printf(
            "\n"
            "SLIDER – Run pairwise comparisons of self-to-self dot plots\n"
            "-i,    --input-file,   Binary dot plot collection\n"
            "-r,    --row-start,    Number of the first record in vertical data range (dot plot)\n"
            "       --row-count,    Total number of rows in vertical data range\n"
            "-c,    --column-start, Number of the first record in horizontal data range (dot plot)\n"
            "       --column-count, Total number of columns in horizontal data range\n"
            "-o,    --outfile,      Output file prefix\n"
            "-t,    --nthreads,     Number of threads to use\n"
            "-h,    --help,         Print this help\n"
    );
}

int main(int argc, char **argv) {
    auto start = std::chrono::steady_clock::now();

    DataRange rowRange = DataRange(DataRange::Role::Rows);
    DataRange colRange = DataRange(DataRange::Role::Columns);

    DotPlotCollection rows;
    DotPlotCollection columns;


    static struct option long_options[] = {
            {"input-file",   required_argument, 0, 'i'},
            {"row-start",    required_argument, 0, 'r'},
            {"row-count",    required_argument, 0, 1000},
            {"column-start", required_argument, 0, 'c'},
            {"column-count", required_argument, 0, 2000},
            {"outfile",      optional_argument, 0, 'o'},
            {"nthreads",     optional_argument, 0, 't'},
            {"help",         no_argument,       0, 'h'},
            {0, 0,                              0, 0}
    };

    printf("*** SLIDER ***\n");
    Utils::print_license();
    Utils::print_citation();
    Utils::print_copyright();

    int long_index = 0, opt;
    while ((opt = getopt_long(argc, argv, "i:r:c:o:t:h",
                              long_options, &long_index)) != -1) {
        switch (opt) {
            case 'i' :
                DotPlotCollection::setDataFileName(optarg);
                break;
            case 'r' :
                rowRange.setStart(std::stoi(optarg));
                break;
            case 1000:
                rowRange.setRange(std::stoi(optarg));
                break;
            case 'c' :
                colRange.setStart(std::stoi(optarg));
                break;
            case 2000:
                colRange.setRange(std::stoi(optarg));
                break;
            case 'o':
                DotPlotCollection::setComparisonMatrixFileNamePrefix(optarg);
                break;
            case 't':
                DotPlotCollection::setNThreads(std::stoi(optarg));
                break;
            case '?':
            case 'h':
                usage();
                exit(EXIT_SUCCESS);
            default:
                fprintf(stderr, "No such option '%c'\n", opt);
                usage();
                exit(EXIT_FAILURE);
        }
    }

    // Print information on proc_binding
    Utils::get_proc_bind_info();

    bool createOutputFile = false;

    DotPlotCollection::readCollectionHeader();
    rowRange.crossValidateWith(colRange);

    rows.setDataRange(&rowRange);
    columns.setDataRange(&colRange);

    rows.populate();
    if (rowRange == colRange) {
        fprintf(stdout, "@@@ Equal ranges given\n");
        columns = rows;
        if (rowRange % colRange) {
            fprintf(stdout, "### This run will create output file.\n");
            createOutputFile = true;
        }
    } else {
        columns.populate();
    }
    auto stop = std::chrono::steady_clock::now();
    auto duration = std::chrono::duration_cast<std::chrono::seconds>(stop - start);
    fprintf(stdout, ">>> [%li s] Data read from input files\n", duration.count());
    fflush(stdout);

    if (createOutputFile)
        DotPlotCollection::prepareComparisonMatrix();
    else
        DotPlotCollection::checkComparisonMatrix();

    int_fast64_t localNJobs = rowRange * colRange,
            totalNJobs = DotPlotCollection::nTotalJobs();

    fprintf(stdout, ">>> We'll run %li jobs which is %12.8f%% of total %li jobs\n", localNJobs,
            localNJobs * 100. / totalNJobs, totalNJobs);

    int localNThreads = DotPlotCollection::getNThreads();
    fprintf(stdout, ">>> Using %d thread%s\n", localNThreads, localNThreads > 1 ? "s" : "");
    fflush(stdout);

    stop = std::chrono::steady_clock::now();
    duration = std::chrono::duration_cast<std::chrono::seconds>(stop - start);
    fprintf(stdout, ">>> [%li s] Run start\n", duration.count());
    fflush(stdout);

    rows.compare(columns);

    stop = std::chrono::steady_clock::now();
    duration = std::chrono::duration_cast<std::chrono::seconds>(stop - start);
    fprintf(stdout, "*** FINISHED in %li s\n", duration.count());

    return 0;
}