#include <iostream>
#include <cstring>
#include <vector>
#include <algorithm>
#include <numeric>
#include <cstdint>
#include <cassert>
#include <mpi.h>
#include <omp.h>
#include <getopt.h>
#include <fcntl.h>
#include <unistd.h>
#include "utils.h"
#include "DotPlotInfoRepository.h"

struct result {
    int_fast64_t jid;   // 8 bytes
    float jaccard;      // 4 bytes
    // 4 bytes of padding added by the compiler for memory alignment

    bool operator<(const result &r) {
        return jid < r.jid;
    }
};

class sparsify_options {
public:
    char *inputFileName = nullptr;
    char *proteinFileName = nullptr;
    char *outABCFileName = nullptr;
    bool bDump = false;
    bool bLargerFirst = false;
    int iRetain = 1400;
    int nThreads = 1;
    float maxInputBuffer = -1;
};

enum timing_s {
    tStart, tInfoStart, tInfoEnd, tDumpStart, tDumpEnd, tProcessingStart, tProcessingEndSortingStart, tSortingEnd,
    tCollectingPivotsStart, tCollectingPivotsEnd, tGatheringStart, tGatheringEnd, tEnd, tNR
};

static void usage() {
    printf(
            "\n"
            "SPARSIFY – Filter file with Jaccard indices and generate input for clustering\n"
            "-i,    --input-raw,            Binary Jaccard index collection (from SLIDER)\n"
            "-p,    --protein-sequences,    Protein sequence database generated by dotrun script\n"
            "-d,    --dump,                 Dump raw data to text file and quit\n"
            "-l,    --larger-first          Put IDs of longer sequences always in the first column\n"
            "-r,    --retain,               Retain this many Jaccard indices per column\n"
            "-o,    --output-abc,           ABC output file\n"
            "-t,    --nthreads,             Number of threads to use\n"
            "-m,    --max-input-buffer,     Maximum amount of RAM [GB] for input buffer\n"
            "-h,    --help,                 Print this help\n"
    );
}

int main(int argc, char **argv) {
    /* MPI START */
    MPI_Init(&argc, &argv);
    int nWorld;
    int rank;
    MPI_Comm_size(MPI_COMM_WORLD, &nWorld);
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    bool isMaster = (rank == 0);
    double localTiming[timing_s::tNR], globalAverageTiming[timing_s::tNR];
    localTiming[tStart] = MPI_Wtime();

    /* Define MPI data type to be able to transfer result structure */
    int blocklengths[] = {1, 1};
    MPI_Datatype types[] = {MPI_LONG_LONG_INT, MPI_FLOAT};
    MPI_Datatype mpi_result;
    MPI_Aint offsets[2];
    offsets[0] = offsetof(result, jid);
    offsets[1] = offsetof(result, jaccard);
    MPI_Type_create_struct(2, blocklengths, offsets, types, &mpi_result);
    MPI_Type_commit(&mpi_result);

    /* I. COMMAND LINE PARSING */

    static struct option long_options[] = {
            {"input-raw",         required_argument, 0, 'i'},
            {"protein-sequences", required_argument, 0, 'p'},
            {"output-abc",        required_argument, 0, 'o'},
            {"dump",              optional_argument, 0, 'd'},
            {"larger-first",      optional_argument, 0, 'l'},
            {"retain",            optional_argument, 0, 'r'},
            {"nthreads",          optional_argument, 0, 't'},
            {"max-input-buffer",  optional_argument, 0, 'm'},
            {"help",              optional_argument, 0, 'h'},
            {0, 0,                                   0, 0}
    };

    printf("*** SPARSIFY ***\n");

    Utils::print_license();
    Utils::print_citation();
    Utils::print_copyright();

    sparsify_options opts;

    int long_index = 0, opt;
    while ((opt = getopt_long(argc, argv, "i:p:o:ldr:t:m:h",
                              long_options, &long_index)) != -1) {
        switch (opt) {
            case 'i':
                opts.inputFileName = optarg;
                break;
            case 'p':
                opts.proteinFileName = optarg;
                break;
            case 'o':
                opts.outABCFileName = optarg;
                break;
            case 'd':
                opts.bDump = true;
                break;
            case 'l':
                opts.bLargerFirst = true;
                break;
            case 'r':
                opts.iRetain = std::stoi(optarg);
                break;
            case 't':
                opts.nThreads = std::stoi(optarg);
                break;
            case 'm':
                opts.maxInputBuffer = 1e9 * std::stof(optarg);
                break;
            case 'h':
                if (isMaster) usage();
                MPI_Finalize();
                exit(EXIT_SUCCESS);
            default:
                if (isMaster) {
                    fprintf(stderr, "No such option %c\n", opt);
                    usage();
                }
                MPI_Abort(MPI_COMM_WORLD, EXIT_FAILURE);
        }
    }

    if (isMaster) {
        Utils::printInfo(argc, argv, nWorld);
    }

    /* II. VALIDATION */
    // Variable needed in different subscopes
    size_t nDotPlots = 0;
    unsigned long long totalABCSize = 0;

    DotPlotInfoRepository info = DotPlotInfoRepository(opts.proteinFileName, DP_SEQL);

    Utils::check_threads(opts.nThreads);
    localTiming[tInfoStart] = MPI_Wtime();
/* III. READ PROTEIN SEQUENCE DATA: EACH LINE CONTAINS ID, LENGTH, NUMBER OF UNIQUE SLIDES AND TOTAL NUMBER OF POINTS*/
    if (isMaster) {
        info.process();
        nDotPlots = info.getNDotPlots();

        if (!opts.bDump && opts.iRetain > nDotPlots) {
            fprintf(stderr, "!!! You want to retain more Jaccard indices per dot plot "
                            "than available (%d > %zu)\n", opts.iRetain, nDotPlots);
            throw std::logic_error("Wrong --retain value");
        }

        if (!opts.bDump && opts.maxInputBuffer < 0) {
            throw std::logic_error("Please explicitly set -m flag!\n");
        }
    }

    MPI_Bcast(&nDotPlots, 1, MPI_UNSIGNED_LONG, 0, MPI_COMM_WORLD);
    MPI_Datatype mpi_dot_plot_set, mpi_dot_plot_wrapper = MPI_DATATYPE_NULL, mpi_dot_plot_first, mpi_dot_plot_regular;
    MPI_Type_contiguous(nDotPlots, MPI_FLOAT, &mpi_dot_plot_set);
    MPI_Type_commit(&mpi_dot_plot_set);

    localTiming[tInfoEnd] = MPI_Wtime();

    /* IV. CHECKING IF THE NUMBER OF SEQUENCES AGREES WITH PROTEIN INFO */

    /* Check if the number of elements found in binary input file
     * corresponds to the number of dot plots in the text file*/
    FILE *fpInfile;
    uint_fast64_t sFromDotPlots, nABCEntries;
    int maxNColumnsRAMLimit = 0;

    if (isMaster) {
        Utils::check_readable(opts.inputFileName);

        uint_fast64_t sInfile = Utils::getFilesize(opts.inputFileName) / sizeof(float);
        sFromDotPlots = nDotPlots * nDotPlots;
        nABCEntries = nDotPlots;
        nABCEntries *= (nDotPlots - 1);
        nABCEntries /= 2;

        if (sInfile != sFromDotPlots) {
            fprintf(stderr, "The number of dot plots from %s (%zu) corresponds to %lu comparisons "
                            "while %s contains %lu\n",
                    opts.proteinFileName, nDotPlots, sFromDotPlots, opts.inputFileName, sInfile);
            throw std::runtime_error("Wrong number of entries either in input file or protein info");
        }
    }

    // one data set occupies this much memory
    int_fast64_t bytesPerColumn = nDotPlots * sizeof(float);
    // how many data sets
    maxNColumnsRAMLimit = opts.maxInputBuffer / bytesPerColumn;

    if (opts.bDump) {
        /* V. PIPELINE A) DUMP THE DATA WITHOUT ANY FILTERING */
        localTiming[tDumpStart] = MPI_Wtime();
        if (isMaster) {
            // 11 = 8 (jaccard) + 2 tab + 1 newline
            totalABCSize += nABCEntries * 11;

            fprintf(stdout, ">>> File %s will occupy %llu bytes\n", opts.outABCFileName, totalABCSize);
            fprintf(stdout,
                    "! WARNING This operation is serial, i.e. only the master process performs reading and writing\n"
                    "! WARNING Please use it for debugging purposes on small test sets\n");
            std::string fnABCDump = Utils::generateFileName(opts.outABCFileName, "dump", "txt");
            FILE *fpDump = fopen64(fnABCDump.c_str(), "wb");
            fpInfile = fopen64(opts.inputFileName, "rb");
            float jaccard;
            for (size_t i = 0; i < nDotPlots; i++) {
                fseeko64(fpInfile, (i + 1) * sizeof(float), SEEK_CUR);
                for (size_t j = i + 1; j < nDotPlots; j++) {
                    if (fread(&jaccard, sizeof(float), 1, fpInfile) != 1) {
                        perror("Incorrect number of bytes read");
                    }
                    if (opts.bLargerFirst) {
                        if (info[i].seqLen >= info[j].seqLen)
                            fprintf(fpDump, "%s\t%s\t%8.6f\n", info[i].id.c_str(), info[j].id.c_str(), jaccard);
                        else
                            fprintf(fpDump, "%s\t%s\t%8.6f\n", info[j].id.c_str(), info[i].id.c_str(), jaccard);
                    } else {
                        fprintf(fpDump, "%s\t%s\t%8.6f\n", info[i].id.c_str(), info[j].id.c_str(), jaccard);
                    }
                }
            }
            fclose(fpInfile);
            off64_t sBytesWritten = ftello64(fpDump);

            fprintf(stdout, ">>> Wrote %li bytes to %s\n", sBytesWritten, fnABCDump.c_str());
            fclose(fpDump);
        }

        localTiming[tDumpEnd] = MPI_Wtime();

    } else {
        /* V. PIPELINE B) FILTER EACH COLUMN LEAVING IRETAIN BEST ENTRIES */
        int *firstColumns = nullptr, *firstColumnCumulative = nullptr,
                *regularColumns = nullptr, *regularColumnCumulative = nullptr, *columnsPerRank = nullptr;

        if (isMaster) {
            fprintf(stdout, "Filtering binary input\n");
            fflush(stdout);

            /* Compute maximum number of elements needed in a localResult array and whether integer type
             * forced by MPI as counts and displacements can hold it */
            int_fast32_t globalMaxNResultsRequired = opts.iRetain * nDotPlots;
            int_fast32_t nWorldMinimum = globalMaxNResultsRequired / INT32_MAX;
            if (globalMaxNResultsRequired % INT32_MAX > 0) ++nWorldMinimum;

            if (nWorld < nWorldMinimum) {
                fprintf(stderr, "!!! Too small world size (%d). "
                                "Please run at least %ld MPI ranks to analyze this amount of data\n",
                        nWorld, nWorldMinimum);
                MPI_Abort(MPI_COMM_WORLD, EXIT_FAILURE);
            }

            fprintf(stdout, "Will read at most %d dot plots during each data retrieval round\n",
                    maxNColumnsRAMLimit);

        }

        MPI_Bcast(&maxNColumnsRAMLimit, 1, MPI_INT, 0, MPI_COMM_WORLD);

        // could be further optimized to give +/- equal reads per session
        int maxNColumnsPerRound = maxNColumnsRAMLimit;

        // number of all jaccard indices that will fit into memory limit
        const int_fast64_t maxNJaccardIndices = nDotPlots * maxNColumnsPerRound;

        int nFirstColumnsSum = 0, nRegularColumnsSum = 0, nRounds = 0;
        if (isMaster) {
            firstColumns = new int[nWorld];
            regularColumns = new int[nWorld];
            columnsPerRank = new int[nWorld];
            firstColumnCumulative = new int[nWorld];
            regularColumnCumulative = new int[nWorld];

            for (int i = 0; i < nWorld; i++) {
                columnsPerRank[i] = (nDotPlots + i) / nWorld;
                firstColumns[i] = columnsPerRank[i] % maxNColumnsPerRound;
                regularColumns[i] = 0; // temporary
            }

            nFirstColumnsSum = std::accumulate(firstColumns, firstColumns + nWorld, 0);
            nRounds = columnsPerRank[nWorld - 1] / maxNColumnsPerRound;

            if (nFirstColumnsSum == 0) {
                for (int i = 0; i < nWorld; i++)
                    regularColumns[i] = columnsPerRank[i] / nRounds;
            } else {
                nRounds += (columnsPerRank[nWorld - 1] % maxNColumnsPerRound > 0 ? 1 : 0);
                if (nRounds > 1)
                    for (int i = 0; i < nWorld; i++)
                        regularColumns[i] = (columnsPerRank[i] - firstColumns[i]) / (nRounds - 1);
            }

            nRegularColumnsSum = std::accumulate(regularColumns, regularColumns + nWorld, 0);

            if (nFirstColumnsSum == 0) {
                for (int i = 0; i < nWorld; i++)
                    firstColumns[i] = regularColumns[i];
                nFirstColumnsSum = nRegularColumnsSum;
            }

            firstColumnCumulative[0] = 0;
            regularColumnCumulative[0] = 0;
            for (int i = 1; i < nWorld; i++) {
                firstColumnCumulative[i] = firstColumnCumulative[i - 1] + firstColumns[i - 1];
                regularColumnCumulative[i] = regularColumnCumulative[i - 1] + regularColumns[i - 1];
            }

        }

        MPI_Bcast(&nRounds, 1, MPI_INT, 0, MPI_COMM_WORLD);
        const int lastRound = nRounds - 1;

        // Another approach...
        // init various column counters
        int nColumnsPerRank = (nDotPlots + rank) / nWorld,
                nColumnsFirstRound = 0, nColumnsFirstCumulative = 0,
                nColumnsRegular = 0, nColumnsRegularCumulative = 0, nColumnsRound = 0,
                currentColumn = 0;

        // scatter
        MPI_Scatter(firstColumns, 1, MPI_INT, &nColumnsFirstRound, 1, MPI_INT, 0, MPI_COMM_WORLD);
        MPI_Scatter(regularColumns, 1, MPI_INT, &nColumnsRegular, 1, MPI_INT, 0, MPI_COMM_WORLD);
        MPI_Scatter(firstColumnCumulative, 1, MPI_INT, &nColumnsFirstCumulative, 1, MPI_INT, 0, MPI_COMM_WORLD);
        MPI_Scatter(regularColumnCumulative, 1, MPI_INT, &nColumnsRegularCumulative, 1, MPI_INT, 0, MPI_COMM_WORLD);

        // broadcast
        MPI_Bcast(&nRegularColumnsSum, 1, MPI_INT, 0, MPI_COMM_WORLD);
        MPI_Bcast(&nFirstColumnsSum, 1, MPI_INT, 0, MPI_COMM_WORLD);

        nColumnsRound = nColumnsFirstRound;
        currentColumn = nColumnsFirstCumulative;

        if (isMaster) {
            delete[] columnsPerRank;
            delete[] firstColumns;
            delete[] regularColumns;
            delete[] firstColumnCumulative;
            delete[] regularColumnCumulative;
        }

        MPI_Offset offsetFirstRound = bytesPerColumn * nColumnsFirstCumulative,
                offsetTotal = bytesPerColumn * nFirstColumnsSum,
                offsetIncrementWorld = bytesPerColumn * nRegularColumnsSum,
                offsetIncrementRegular = bytesPerColumn * nColumnsRegularCumulative,
                currentOffset = offsetFirstRound;

        if (isMaster) {
            fprintf(stdout, "General information:\n"
                            "nRounds = %d\n"
                            "maxNColumnsRAMLimit = %d, maxNJaccardIndices = %li\n"
                            "First column sum = %d, Regular column sum = %d\n"
                            "offsetTotal = %llu, offsetIncrementWorld = %llu\n"
                            "--------------------------------------------------------\n",
                    nRounds, maxNColumnsRAMLimit, maxNJaccardIndices, nFirstColumnsSum, nRegularColumnsSum, offsetTotal,
                    offsetIncrementWorld);
            fflush(stdout);
        }

        usleep(rank * 10000);
        fprintf(stdout, "Node %d: nColumnsPerRank = %d\n"
                        "nColumnsFirstRound = %d, nColumnsFirstCumulative = %d\n"
                        "nColumnsRegular = %d, nColumnsRegularCumulative = %d\n"
                        "offsetFirstRound = %llu\n"
                        "--------------------------------------------------------\n",
                rank, nColumnsPerRank, nColumnsFirstRound, nColumnsFirstCumulative, nColumnsRegular,
                nColumnsRegularCumulative, offsetFirstRound);
        fflush(stdout);

        /* Avoid overflow; nLocalResults = nColumnsPerRank * opts.iRetain */
        int_fast64_t nFilteredLocalResults = nColumnsPerRank;
        nFilteredLocalResults *= opts.iRetain;

        // Buffer for raw data read by MPI IO
        float *pJaccardBuffer = new float[maxNJaccardIndices];
        for (int_fast64_t j = 0; j < maxNJaccardIndices; j++)
            pJaccardBuffer[j] = 0.0f;

        int_fast64_t nLocalResultBuffer = nDotPlots * opts.nThreads;
        result *pLocalResultBuffer = new result[nLocalResultBuffer];
        for (int_fast64_t j = 0; j < nLocalResultBuffer; j++) {
            pLocalResultBuffer[j].jaccard = 0.0f;
            pLocalResultBuffer[j].jid = -1;
        }

        result *pFilteredLocalResults = new result[nFilteredLocalResults];
        for (int_fast64_t j = 0; j < nFilteredLocalResults; j++) {
            pFilteredLocalResults[j].jaccard = 0.0f;
            pFilteredLocalResults[j].jid = -1;
        }

        MPI_Type_contiguous(nColumnsFirstRound, mpi_dot_plot_set, &mpi_dot_plot_first);
        MPI_Type_commit(&mpi_dot_plot_first);
        MPI_Type_contiguous(nColumnsRegular, mpi_dot_plot_set, &mpi_dot_plot_regular);
        MPI_Type_commit(&mpi_dot_plot_regular);
        mpi_dot_plot_wrapper = mpi_dot_plot_first;

        MPI_File fh;
        MPI_Status status;

        MPI_CHECK(MPI_File_open(MPI_COMM_WORLD, opts.inputFileName, MPI_MODE_RDONLY, MPI_INFO_NULL, &fh),
                  "Could not read the input symmetric matrix");

        double parStart = MPI_Wtime();
        int nColumnsProcessed = 0;
        double localReadTime = 0, localReadStartTime, localReadStopTime, currentReadTime, currentReadRate;

        localTiming[tProcessingStart] = MPI_Wtime();

        /* MAIN LOOP */
        for (uint iRound = 0; iRound < nRounds; iRound++) {

            localReadStartTime = MPI_Wtime();
            // READ DATA -> collective MPI call
            MPI_CHECK(MPI_File_read_at_all(fh, currentOffset, pJaccardBuffer, 1, mpi_dot_plot_wrapper,
                                           &status),
                      "Could not read data series");
            localReadStopTime = MPI_Wtime();
            currentReadTime = localReadStopTime - localReadStartTime;
            localReadTime += currentReadTime;
            if (currentReadTime > 0) {
                currentReadRate = bytesPerColumn * nColumnsRound / currentReadTime;

                double wRLog = std::log10(currentReadRate);
                const char *prefix[] = {"", "k", "M", "G", "T"};
                int prefixIndex = int(wRLog) / 3;
                if (prefixIndex > 3)
                    prefixIndex = 3;
                int wRDivider = prefixIndex * 3;
                if (wRDivider > 0)
                    currentReadRate /= std::pow(10, wRDivider);

                fprintf(stdout,
                        "-> Node %d: round %d: read %u columns at offset %lld (column %d), end offset %lld: "
                        "rate %5.3f %sB/s\n",
                        rank, iRound, nColumnsRound, currentOffset, currentColumn, offsetTotal,
                        currentReadRate, prefix[prefixIndex]);
            } else {
                fprintf(stdout,
                        "-> Node %d: round %d: read %u columns at offset %lld (column %d), end offset %lld: "
                        "rate inf B/s\n",
                        rank, iRound, nColumnsRound, currentOffset, currentColumn, offsetTotal);
            }

            fflush(stdout);
            // omp pragma for loop for processing
            // -> for each set read
#pragma omp parallel default(none) shared(pJaccardBuffer, nDotPlots, opts, nColumnsRound, currentColumn, iRound, \
    pLocalResultBuffer, pFilteredLocalResults, rank, stdout, nColumnsProcessed, nColumnsPerRank)
            {
                int ompThreadID = omp_get_thread_num();
                size_t locResIndex = ompThreadID * nDotPlots; // for pLocalResultBuffer
                result *pLocalResultBufferBegin = pLocalResultBuffer + locResIndex;
                size_t iResults;

#pragma omp for schedule(dynamic)
                for (uint iDotPlot = 0; iDotPlot < nColumnsRound; iDotPlot++) {
                    size_t jIndex = iDotPlot * nDotPlots; // for pJaccardBuffer
                    // for pFilteredLocalResults // this should depend on iRound
                    size_t filtResIndex = (nColumnsProcessed + iDotPlot);
                    filtResIndex *= opts.iRetain;
                    result *pLocalResultBufferEnd = pLocalResultBufferBegin;
                    // pointer to this thread's buffer start
                    float *pLocalJaccardBuffer = pJaccardBuffer + jIndex;

                    size_t j;
                    size_t i = currentColumn + iDotPlot;
                    size_t index = i; // job id
                    // -> write jaccards along with calculated job ids to a temporary per-thread result buffer
                    for (j = 0; j < i; j++) {
                        if (pLocalJaccardBuffer[j] > 0.0f) {
                            pLocalResultBufferEnd->jaccard = pLocalJaccardBuffer[j];
                            pLocalResultBufferEnd->jid = index;
                            ++pLocalResultBufferEnd;
                        }
                        index += nDotPlots;
                    }

                    index = i * nDotPlots + i + 1;
                    for (j = i + 1; j < nDotPlots; j++) {
                        if (pLocalJaccardBuffer[j] > 0.0f) {
                            pLocalResultBufferEnd->jaccard = pLocalJaccardBuffer[j];
                            pLocalResultBufferEnd->jid = index;
                            ++pLocalResultBufferEnd;
                        }
                        ++index;
                    }

                    iResults = pLocalResultBufferEnd - pLocalResultBufferBegin;

                    // -> sort buffer by jaccard value
                    std::sort(pLocalResultBufferBegin, pLocalResultBufferEnd,
                              [](const result &l, const result &r) -> bool { return l.jaccard > r.jaccard; });

                    // -> determine how much we keep – iRetain as cutoff
                    size_t k = opts.iRetain > iResults ? iResults : opts.iRetain;

                    // -> copy values to the final result buffer of the node
                    for (int_fast64_t m = filtResIndex, n = 0; n < k; m++, n++) {
                        // copy
                        pFilteredLocalResults[m].jid = pLocalResultBufferBegin[n].jid;
                        pFilteredLocalResults[m].jaccard = pLocalResultBufferBegin[n].jaccard;
                        // clear
                        pLocalResultBufferBegin[n].jid = -1;
                        pLocalResultBufferBegin[n].jaccard = 0.0f;
                    }

                } /* END omp for */

            } /* END omp parallel section */


            // update start and offset for the next read operation
            if (iRound < lastRound) {

                nColumnsProcessed += nColumnsRound;
                if (iRound == 0) {
                    // update offset
                    nColumnsRound = nColumnsRegular;
                    currentColumn = nFirstColumnsSum + nColumnsRegularCumulative;
                    mpi_dot_plot_wrapper = mpi_dot_plot_regular;
                } else {
                    currentColumn += nRegularColumnsSum;
                }

                currentOffset = offsetTotal + offsetIncrementRegular;
                offsetTotal += offsetIncrementWorld;
            }

        } /* END OF DATA RETRIEVAL LOOP */

        MPI_CHECK(MPI_File_close(&fh), "Could not close binary symmetric matrix");

        // delete two buffers not needed anymore
        delete[] pJaccardBuffer;
        delete[] pLocalResultBuffer;

        fprintf(stdout, "Parallel section on rank %d took %.1f s\n", rank, MPI_Wtime() - parStart);
        fflush(stdout);
        MPI_Barrier(MPI_COMM_WORLD);

        localTiming[tProcessingEndSortingStart] = MPI_Wtime();

        // NOTE we keep the pFilteredLocalResults -> it is going to be used later on
        std::sort(pFilteredLocalResults, pFilteredLocalResults + nFilteredLocalResults);
        localTiming[tSortingEnd] = MPI_Wtime();

        // calculate local pivot points, i.e. job ids which constitute batch limits for later slave->master communication
        /* Pivots section
         * localResults is divided into iRetain sections with pivot values picked from the entire
         * population as their borders */
        // default: nPivots approx.
        const int nPivots = opts.iRetain;
        int nGlobalPivotCandidates = nPivots * nWorld;
        int_fast64_t *localPivots = new int_fast64_t[nPivots];
        int_fast64_t *globalPivotsSelection = nullptr;
        int_fast64_t *globalPivots = new int_fast64_t[nPivots];
        if (isMaster) {
            globalPivotsSelection = new int_fast64_t[nGlobalPivotCandidates];
        }

        // Returns an iterator pointing to the first element in the range [first, last) that is not less than value.
        // This is here in case we had less non-zero Jaccard indices than nPivots value which is when we have -1 job id
        result zeroResult;
        zeroResult.jaccard = 0.0f;
        zeroResult.jid = 0;
        result *lowestPos = std::lower_bound(pFilteredLocalResults, pFilteredLocalResults + nFilteredLocalResults,
                                             zeroResult);
        int_fast64_t sum_zero = std::distance(pFilteredLocalResults, lowestPos), sum = sum_zero;
        for (int k = 0; k < nPivots; k++) {
            localPivots[k] = pFilteredLocalResults[sum].jid;
            sum += (nFilteredLocalResults + k - sum_zero) / nPivots;
        }

        localTiming[tCollectingPivotsStart] = MPI_Wtime();

        // communicate local pivot points to master
        MPI_Gather(localPivots, nPivots, MPI_LONG_LONG_INT,
                   globalPivotsSelection, nPivots, MPI_LONG_LONG_INT, 0, MPI_COMM_WORLD);
        localTiming[tCollectingPivotsEnd] = MPI_Wtime();

        // MASTER ONLY: determine global pivot points
        if (isMaster) {
            /* SORT CANDIDATES */
            std::sort(globalPivotsSelection, globalPivotsSelection + nGlobalPivotCandidates);

            sum = 0;
            for (int k = 0; k < nPivots; k++) {
                globalPivots[k] = globalPivotsSelection[sum];
                sum += (nGlobalPivotCandidates + k) / nPivots;
            }
        }

        /* Announce pivots to the world */
        MPI_Bcast(globalPivots, nPivots, MPI_LONG_LONG_INT, 0, MPI_COMM_WORLD);

        // compute section sizes
        /* Initiate count and displacement arrays needed by MPI_Gather */
        int *localCounts = new int[nPivots];
        int *localDisplacements = new int[nPivots];
        int *globalCountSum = nullptr,
                *displacements = nullptr,
                *batchCounts = nullptr,
                *globalCounts = nullptr;
        int nGlobalCounts = nPivots * nWorld;
        if (isMaster) {
            globalCountSum = new int[nPivots];
            globalCounts = new int[nGlobalCounts];
            displacements = new int[nWorld];
            batchCounts = new int[nWorld];
        }

        result *prev, *current;
        result sectionLimit;
        sectionLimit.jaccard = 0.0f;
        sectionLimit.jid = 0;
        sum = 0;
        for (int i = 0; i < nPivots; i++) {
            if (i == 0)
                prev = pFilteredLocalResults;
            else
                prev = current;

            if (i == nPivots - 1)
                current = pFilteredLocalResults + nFilteredLocalResults;
            else {
                sectionLimit.jid = globalPivots[i + 1];
                current = std::lower_bound(prev, pFilteredLocalResults + nFilteredLocalResults, sectionLimit);
            }

            localCounts[i] = static_cast<int>(std::distance(prev, current));
            sum += localCounts[i];
        }
        assert(sum == nPivots * nColumnsPerRank);

        localDisplacements[0] = 0;
        for (int i = 1; i < nPivots; i++) {
            localDisplacements[i] = localDisplacements[i - 1] + localCounts[i - 1];
        }

        // master sums the sizes of every batch across the nodes, finds maximum and uses it to allocate a buffer for results
        // just a check if the numbers match iRetain * nDotPlots
        int_fast64_t nGlobalResults;
        MPI_Reduce(&nFilteredLocalResults, &nGlobalResults, 1, MPI_LONG_LONG_INT, MPI_SUM, 0, MPI_COMM_WORLD);

        // sum of elements per batch
        MPI_Reduce(localCounts, globalCountSum, nPivots, MPI_INT, MPI_SUM, 0, MPI_COMM_WORLD);

        // gather all counts
        MPI_Gather(localCounts, nPivots, MPI_INT, globalCounts, nPivots,
                   MPI_INT, 0, MPI_COMM_WORLD);

        result *batchResults = nullptr;
        result *shiftLocalResults = nullptr;
        FILE *fpABCOut = nullptr;

        // for each batch gather results, sort them and write non-redundant entries to ABC file
        if (isMaster) {
            fprintf(stdout, "Gathering and writing results\n");
            int maxReceived = *std::max_element(globalCountSum, globalCountSum + nPivots);
            assert(nGlobalResults == nPivots * nDotPlots);
            fprintf(stdout, "*** The global number of results stored is %li and max count per batch is %i\n",
                    nGlobalResults, maxReceived);

            // we don't get from ourselves, hence the number of elements
            batchResults = new result[maxReceived];

            fpABCOut = fopen64(Utils::generateFileName(opts.outABCFileName, "filtered", "txt").c_str(), "wb");
        }

        int totalCountPerBatch = 0;
        int_fast64_t row = 0, col = 0;
        int_fast64_t lastJobID;

        uint *occurences = NULL;
        if (isMaster) {
            /* A fuse: check which columns are not present, due to being all zeroes */
            occurences = new uint[nDotPlots];
            for (int j = 0; j < nDotPlots; ++j) {
                occurences[j] = 0;
            }
        }

        localTiming[tGatheringStart] = MPI_Wtime();
        // for each of nPivots packets of different length, determined for each node
        for (int iBatch = 0; iBatch < nPivots; iBatch++) {
            if (isMaster) {
                totalCountPerBatch = 0;
                for (int r = 0; r < nWorld; r++) {
                    batchCounts[r] = globalCounts[r * nPivots + iBatch];
                    if (r == 0)
                        displacements[r] = 0;
                    else
                        displacements[r] = displacements[r - 1] + batchCounts[r - 1];
                    totalCountPerBatch += batchCounts[r];
                }
                assert(totalCountPerBatch == globalCountSum[iBatch]);
                if (iBatch % 100 == 0) {
                    fprintf(stdout, "\rGathering round %d\n", iBatch);
                    fflush(stdout);
                }
            }

            shiftLocalResults = pFilteredLocalResults + localDisplacements[iBatch];

            /* Use custom data type created earlier */
            MPI_Gatherv(shiftLocalResults, localCounts[iBatch], mpi_result,
                        batchResults, batchCounts, displacements, mpi_result, 0, MPI_COMM_WORLD);

            if (isMaster) {
                std::sort(batchResults, batchResults + totalCountPerBatch);
                lastJobID = -1;

                for (int i = 0; i < totalCountPerBatch; ++i) {
                    if (batchResults[i].jid == -1 || batchResults[i].jid == lastJobID)
                        continue;
                    else if (batchResults[i].jid < 0)
                        throw std::domain_error("Negative result");
                    else {

                        row = batchResults[i].jid / nDotPlots;
                        col = batchResults[i].jid % nDotPlots;
                        if (row > col) {
                            throw std::runtime_error("Row index is greater than column index");
                        }

                        ++occurences[row];
                        ++occurences[col];
                        if (opts.bLargerFirst) {
                            if (info[row].seqLen >= info[col].seqLen)
                                fprintf(fpABCOut, "%s\t%s\t%8.6f\n",
                                        info[row].id.c_str(), info[col].id.c_str(), batchResults[i].jaccard);
                            else
                                fprintf(fpABCOut, "%s\t%s\t%8.6f\n",
                                        info[col].id.c_str(), info[row].id.c_str(), batchResults[i].jaccard);
                        } else {
                            fprintf(fpABCOut, "%s\t%s\t%8.6f\n", info[row].id.c_str(), info[col].id.c_str(),
                                    batchResults[i].jaccard);
                        }
                        lastJobID = batchResults[i].jid;
                    }
                } /* LOOP OVER RESULTS */
            } /* MASTER */
        } /* BATCH LOOP */

        MPI_Barrier(MPI_COMM_WORLD);
        localTiming[tGatheringEnd] = MPI_Wtime();

        if (isMaster) {
            uint nExcludedDotPlots = 0;
            std::vector<std::string> exclusionsThatShouldNotHaveHappened;
            for (int j = 0; j < nDotPlots; ++j) {
                if (occurences[j] == 0) {
                    ++nExcludedDotPlots;
                    if (info[j].nPoints > 0)
                        exclusionsThatShouldNotHaveHappened.emplace_back(info[j].id);
                }
            }
            if (nExcludedDotPlots > 0)
                fprintf(stdout,
                        "WARNING There were %u excluded sequences due to having 0 points on their dot plots.\n"
                        "You should not include them in the calculations for efficiency reasons.\n",
                        nExcludedDotPlots);

            if (exclusionsThatShouldNotHaveHappened.size() > 0) {
                fprintf(stdout,
                        "There were sequences that were excluded although they have points on their dot plots."
                        " This may be worth investigating. The list is below:\n");
                for (auto sequence : exclusionsThatShouldNotHaveHappened) {
                    std::cout << sequence << std::endl;
                }
            }
            fprintf(stdout, "(Hip)MCL should find %zu vertices.\n", nDotPlots - nExcludedDotPlots);
            delete[] occurences;
        } /* IF MASTER */


        if (isMaster) {
            fclose(fpABCOut);
            delete[] globalPivotsSelection;
            delete[] globalCounts;
            delete[] globalCountSum;
            delete[] displacements;
            delete[] batchCounts;
            delete[] batchResults;
        }
        delete[] pFilteredLocalResults;
        delete[] localPivots;
        delete[] localDisplacements;
        delete[] localCounts;
        delete[] globalPivots;
    } /* END IF */

    MPI_Type_free(&mpi_result);
    if (!opts.bDump) {
        MPI_Type_free(&mpi_dot_plot_first);
        MPI_Type_free(&mpi_dot_plot_regular);
    }
    MPI_Type_free(&mpi_dot_plot_set);
    localTiming[tEnd] = MPI_Wtime();

    if (isMaster) {
        fprintf(stdout, "@@@ TIMINGS ON MASTER @@@\n"
                        "Reading protein information took %.1f seconds\n",
                localTiming[tInfoEnd] - localTiming[tInfoStart]);
        if (opts.bDump)
            fprintf(stdout, "Dumping data to ABC text file took %.1f seconds\n",
                    localTiming[tDumpEnd] - localTiming[tDumpStart]);
        else {
            fprintf(stdout, "Preparations for processing input took %.1f seconds\n",
                    localTiming[tProcessingStart] - localTiming[tInfoEnd]);
            fprintf(stdout, "Processing binary input file took %.1f seconds\n",
                    localTiming[tProcessingEndSortingStart] - localTiming[tProcessingStart]);
            fprintf(stdout, "Sorting data took %.1f seconds\n",
                    localTiming[tSortingEnd] - localTiming[tProcessingEndSortingStart]);
            fprintf(stdout, "Generating data for the final gathering took %.1f seconds\n",
                    localTiming[tGatheringStart] - localTiming[tSortingEnd]);
            fprintf(stdout, "\tCollecting pivots subsection took %.1f seconds\n",
                    localTiming[tCollectingPivotsEnd] - localTiming[tCollectingPivotsStart]);
            fprintf(stdout, "Gathering and writing data to ABC text file took %.1f seconds\n",
                    localTiming[tGatheringEnd] - localTiming[tGatheringStart]);
        }
        fprintf(stdout, "*** FINISHED in %.1f s\n", localTiming[tEnd] - localTiming[tStart]);
    }

    MPI_Finalize();
    return 0;
}
